> module Evaluate
> where

> import Prelude hiding (($))
> import Data.Char
> import Data.Maybe
> import Data.List (sort)
> import System.IO
> import Control.Monad (mplus, msum, liftM2)
> import Data.Array
> import Data.List (genericLength)
> import qualified System.Console.Readline as Readline

> import PrettierNG

> import Options
> import Core

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Values}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> data Val
>   =  VBool   !Bool
>   |  VNat    !Integer
>   |  VString !String
>   |  VType -- type argument marker
>   |  Closure Pat Expr Bool !Env -- recursive environment? 
>   |  VRecord ![Row Val]
>   |  VCon    !Con !Val
>   |  Addr    !Integer
>   |  Cont    !Stack
>   |  VStruct     !Env -- first-class environments
>   |  VClass Expr Env
>   |  VObject !Env -- objects
>   |  VSuper  !Env -- objects
>   |  VMethod Expr !Env
>   |  VArray !(Array Integer Val)

To create a sorted/normalized record value after evaluation of the
components:

> vrecord rs = VRecord (sort rs)

NB. `sort' must be a stable sorting algorithm, so that the relative
order of numbered fields is not changed.

Die Konstruktoren sind strikt in allen Argumenten (damit
z.B. arithmetische Operationen auch wirklich ausgefhrt werden).

> ppVal (VBool b) 
>   =  ppBool b
> ppVal (VNat n) 
>   =  pretty n
> ppVal (VString s) 
>   =  pretty s
> ppVal VType
>   =  text "<>"
> ppVal (Closure p e recursive env)
>   =  parens (align "\\ " (ppPat p <> text " =>" <+> ppExpr e) <+>
>              if recursive then text "[..]"
>                           else ppEnv' env) -- HACK
> ppVal (VRecord rs)
>   =  parens (group (intersperse (char ',' <> sp) (map (ppRow ppVal) rs)))

> ppVal (VCon c v) 
>   =  parens (text c <+> ppVal v)
> ppVal (Addr a) 
>   =  text "&" <> pretty a
> ppVal (Cont ctl)
>   =  ppStack ctl
> ppVal (VStruct env)
>   =  group (text "struct" <+> ppEnv env)

> ppVal (VClass e env)
>   =  group (text "class" <+> ppExpr e <+> ppEnv env)
> ppVal (VObject env)
>   =  group (text "object" <+> ppEnv env)
> ppVal (VSuper env)
>   =  group (text "super" <+> ppEnv env)
> ppVal (VMethod e env)
>   =  group (text "method" <+> ppExpr e <+> ppEnv env)
> ppVal (VArray a)
>   =  ppList ppVal (elems a)

 ppVal (VMethod p e env)
   =  parens (align "method " (ppPat p <> text " =>" <+> ppExpr e) <+> ppEnv' env)

 ppVal e -- catch-all (vorlufig)
   =  pretty e

> instance Show Val where
>   showsPrec _ (VBool b) = shows b
>   showsPrec _ (VNat n) = shows n
>   showsPrec _ (VString s) = shows s
>   showsPrec _ (Closure x e _ _)
>     = showString "<" . shows x . showString ", " . shows e . showString ", [..]>"
>   showsPrec _ (VRecord fs)
>     =  shows fs
>   showsPrec _ (VCon c v) 
>     = showString "(" . showString c .showString " " . shows v . showString ")"
>   showsPrec _ (Cont ctl)
>     = showString "<" . shows ctl . showString ">"

> instance Pretty Val

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Environment}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> type Env
>   =  [(Var, Val)]

> ppEnv' env
>   =  text "[..]" -- ppList ppBind env
> ppEnv env
>   =  ppList ppBind env

> ppBind (x, v)
>   =  align (x ++ " |-> ") (ppVal v)

 align s d = block (length s) (text s <> d)

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Stack frames}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> type Stack
>   =  [Frame]

> data Frame
>   =  If1 {--} Expr Expr Env
>   |  UnOp1 UnOp {--}
>   |  BinOp1 {--} BinOp Expr Env
>   |  BinOp2 Val BinOp {--}
>   |  Let1 Pat {--} Expr Env
>   |  LetOpen1 {--} Expr Env
>   |  LetInherit1 (Maybe Var) {--} Expr Env
>   |  App1 {--} Expr Env
>   |  App2 Val  {--}
>   |  MApp2 Val  {--} Env -- Variante von `App2': `Env' wird fr die Implementierung direkter Methodenaufrufe bentigt
>   |  TApp2 {--} Val -- noch eine Variante von `App2': beim Einfuegen eines Typarguments ist das Funktionsargument anschliessend schon ausgewertet
>   |  Recordi [Row Val] (Row ()) {--} [Row Expr] Env
>   |  Recordn [Row Val] (Row ()) {--}
>   |  Sel1 {--} Lab Env -- `Env' wird fr die Implementierung von `super' bentigt

>   |  Con1 Con {--}
>   |  Seq1 {--} Expr Env
>   |  Ref1 {--}
>   |  Assign1 {--} Expr Env
>   |  Assign2 Val {--}
>   |  Deref1 {--}
>   |  Case1 {--} [(Pat, Expr)] Env
>   |  Continue1 {--} Expr Env
>   |  Continue2 Val {--}


>   |  Throw1 {--}
>   |  Try1 {--} [(Pat, Expr)] Env
>   |  Call1 Prim {--}
> --  |  Show1 {--}
>   |  While1 Expr Expr Env

>   |  Lookup1 {--} Var

>   |  New1 {--}
>   |  Inherit1 {--}
>   |  InheritIn1 Var {--} Expr Env
>   |  InheritIn2 Var {--} Expr Env
>   |  InheritIn3 Var Val {--}
>   |  Append1 {--} Expr Env
>   |  Append2 Val {--}

>   |  ArrFun1 {--} Var Expr Env
>   |  ArrFun2i !Integer !Integer [Val] Var Expr Env
>   |  ArrListi [Val] {--} [Expr] Env
>   |  ArrListn [Val] {--}
>   |  Subscript1 {--} Expr Env
>   |  Subscript2 Val {--}

>      deriving (Show)

> instance Pretty Frame

> hole = text "o"

> ppFrame (If1 {--} e2 e3 env)
>   =  group (text "if " <> 
>             block 3 (hole <+> block 5 (text "then " <> ppExpr e2) <+>
>                               block 5 (text "else " <> ppExpr e3)) <+> ppEnv env)
> ppFrame (UnOp1 op {--})
>   =  group (pretty op <+> hole)
> ppFrame (BinOp1 {--} op e env)
>   =  group (hole <> text " " <> pretty op <+> block 2 (ppExpr e) <+> ppEnv env)
> ppFrame (BinOp2 v op {--})
>   =  group (ppVal v <> text " " <> pretty op <+> block 2 hole)
> ppFrame (Let1 p {--} e env)
>   =  group (text "let " <> block 4 (ppPat p <+> text "=" <+> hole) <> 
>             block 1 (sp <> text "in " <> block 3 (ppExpr e) <+> ppEnv env))
> ppFrame (App1 {--} e env)
>   =  group (hole <> text " @" <+> ppExpr e <+> ppEnv env)
> ppFrame (App2 v {--})
>   =  group (ppVal v <> text " @" <+> hole)
> ppFrame (MApp2 v {--} env)
>   =  group (ppVal v <> text " @" <+> hole <+> ppEnv env)
> ppFrame (TApp2 {--} v)
>   =  group (hole <> text " @" <+> ppVal v)
> ppFrame (Recordi rs1 r rs2 env)
>   =  group (parens (group (intersperse (char ',' <> sp) 
>               (map (ppRow ppVal) rs1 ++ [ppRow (const hole) r] ++ map (ppRow ppExpr) rs2))) <+> 
>             ppEnv env)
> ppFrame (Recordn rs1 r)
>   =  parens (group (intersperse (char ',' <> sp) 
>         (map (ppRow ppVal) rs1 ++ [ppRow (const hole) r])))
> ppFrame (Sel1 {--} l env)
>   =  group (hole <> text "." <> ppLab l)

> ppFrame (Con1 c {--})
>   =  group (text c <+> hole)
> ppFrame (Seq1 {--} e env)
>   =  group (hole <> text ";" <+> ppExpr e <+> ppEnv env)
> ppFrame (Ref1 {--})
>   =  group (text "ref" <+> hole)
> ppFrame (Assign1 {--} e env)
>   =  group (hole <+> text ":=" <+> ppExpr e <+> ppEnv env)
> ppFrame (Assign2 v {--})
>   =  group (ppVal v <+> text ":=" <+> hole)
> ppFrame (Deref1 {--})
>   =  group (text "!" <> hole)
> ppFrame (Case1 {--} rs env)
>   =  group (text "case " <> hole <+> text "of " <>
>        intersperse (sp <> text " | ") (map ppRule rs) <+> text "end" <+> ppEnv env)
> ppFrame (Continue1 {--} e env)
>   =  group (text "continue" <+> hole <+> ppExpr e <+> ppEnv env)
> ppFrame (Continue2 v {--})
>   =  group (text "continue" <+> ppVal v <+> hole)
> ppFrame (Try1 {--} rs env)
>   =  group (text "try" <> block 2 (sp <> hole) <+> text "catch" <+>
>        block 2 (intersperse (sp <> text " | ") (map ppRule rs)) <+> text "end" <+> ppEnv env)
> ppFrame (Throw1 {--})
>   =  group (text "throw" <+> hole)
> ppFrame (Call1 f {--})
>   =  group (pretty f <+> hole)
> --ppFrame (Show1 {--})
> --  =  group (text "show" <+> hole)
> ppFrame (While1 e1 e2 env)
>   =  group (text "while " <> 
>             block 2 (ppExpr e1 <> text " do" <+> ppExpr e2) <+> ppEnv env)
> ppFrame (LetOpen1 {--} e env)
>   =  group (text "let open " <> block 9 (hole) <> 
>             block 1 (sp <> text "in " <> block 3 (ppExpr e) <+> ppEnv env))
> ppFrame (LetInherit1 Nothing {--} e env)
>   =  group (text "let inherit " <> block 9 (hole) <> 
>             block 1 (sp <> text "in " <> block 3 (ppExpr e) <+> ppEnv env))
> ppFrame (LetInherit1 (Just x) {--} e env)
>   =  group (text "let inherit " <> text x <+> text "=" <+> block 9 (hole) <> 
>             block 1 (sp <> text "in " <> block 3 (ppExpr e) <+> ppEnv env))
> ppFrame (Lookup1 {--} x)
>   =  group (hole <> text"$" <> text x)

> ppFrame (New1 {--})
>   =  group (text "new " <> hole)
> ppFrame (Append1 {--} e env)
>   =  group (hole <+> ppExpr e <+> ppEnv env)
> ppFrame (Append2 v {--})
>   =  group (ppVal v <+> hole)
> ppFrame (InheritIn1 x {--} e2 env)
>   =  group (text "inherit " <> block 8 (text x <> text " =" <+> hole) <> 
>             block 1 (sp <> text "in " <> block 3 (ppExpr e2) <+> text "end") <+> ppEnv env)
> ppFrame (InheritIn2 x {--} e2 env)
>   =  group (text "inherit " <> block 8 (text x <> text " =" <+> hole) <> 
>             block 1 (sp <> text "in " <> block 3 (ppExpr e2) <+> text "end") <+> ppEnv env)
> ppFrame (InheritIn3 x v1 {--})
>   =  group (text "inherit " <> block 8 (text x <> text " =" <+> ppVal v1) <> 
>             block 1 (sp <> text "in " <> block 3 (hole) <+> text "end"))

> ppFrame (ArrFun1 {--} x e2 env)
>   =  parens (align "array " (brackets hole <+> text x <> text " =>" <+> ppExpr e2) <+> ppEnv env)
> ppFrame (ArrFun2i i n vs x e2 env)
>   =  parens (brackets (intersperse (char ',' <> sp) (map ppVal vs ++ [hole])) <+> text "+" <+>
>              align "array " (text ("[" ++ show i ++ ".." ++ show (n-1) ++ "]") <+> text x <> text " =>" <+> ppExpr e2) <+> ppEnv env)
> ppFrame (ArrListi vs es env)
>   =  brackets (intersperse (char ',' <> sp) (map ppVal vs ++ [hole] ++ map ppExpr es) <+> ppEnv env)
> ppFrame (ArrListn vs)
>   =  brackets (intersperse (char ',' <> sp) (map ppVal vs ++ [hole]))
> ppFrame (Subscript1 {--} e2 env)
>   =  group (hole <> brackets (ppExpr e2) <+> ppEnv env)
> ppFrame (Subscript2 v1 {--})
>   =  group (ppVal v1 <> brackets hole)

 ppFrame (Force1 {--})
   =  group (text "force " <> hole)

 ppFrame f -- catch-all (vorlufig)
   =  pretty f

> recordi fs1 l []  env  =  Recordn fs1 l
> recordi fs1 l fs2 env  =  Recordi fs1 l fs2 env

> arrFun2i i n vs x e2 env
>   | i < n      =  ArrFun2i i n vs x e2 env
>   | otherwise  =  ArrListn vs

> arrListi vs [] env  =  ArrListn vs
> arrListi vs es env  =  ArrListi vs es env

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Memory}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> type Loc
>   =  Integer
> type Mem
>   =  (Loc, [(Loc, Val)]) -- nchste freie Location

> ppMem (_, mem)
>   =  ppList ppCell mem

> ppCell (a, v)
>   =  group (text "&" <> pretty a <> text ":" <+> ppVal v)

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Instructions}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> data State
>   =   Stop Val Mem
>   |   Forward  Expr Env Stack Mem
>   |   Backward Val Stack Mem
>   |   CoreDumped Doc


> {-# INLINE xforward #-}
> xforward e env ctl mem
>   =  return (Forward e env ctl mem)
> {-# INLINE xbackward #-}
> xbackward v ctl mem
>   =  return (Backward v ctl mem)
> {-# INLINE coreDumped #-}
> coreDumped s
>   =  return (CoreDumped s)

> runtimeTypeError name tp arg
>   =  coreDumped (text "`" <> text name <> text "' expects " <> text tp <> text ":" <> nl <>
>        (group (text name <+> ppVal arg)))


> instance Show State where
>   showsPrec _ (Stop v mem) = shows v . showString " | " . shows mem
>   showsPrec _ (Forward  e env ctl mem) 
>     = shows e . showString " " . shows env . showString " >> " . shows ctl . showString " | " . shows mem
>   showsPrec _ (Backward v ctl mem) = shows v . showString " << " . shows ctl . showString " | " . shows mem
>   showsPrec _ (CoreDumped d)  = showString ("core dumped: " ++ render (Page 80) d)

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{The abstract machine}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

One step of the abstract machine.

> {-# INLINE step #-}
> step :: State -> IO State

> step (Stop v mem)
>   =  return (Stop v mem)
> step (Forward e env ctl mem)
>   =  forward e env ctl mem
> step (Backward v [] mem)
>   =  return (Stop v mem)
> step (Backward v (f : fs) mem)
>   =  backward f v fs mem
> step (CoreDumped s)
>   =  coreDumped s

Forward step.

> forward :: Expr -> Env -> Stack -> Mem -> IO State
> forward (Bool b) env ctl mem
>   =  xbackward (VBool b) ctl mem
> forward (If e1 e2 e3) env ctl mem
>   =  xforward e1 env (If1 {--} e2 e3 env : ctl) mem
> forward (Nat n) env ctl mem
>   =  xbackward (VNat n) ctl mem
> forward (UnOp op e) env ctl mem
>   =  xforward e env (UnOp1 op {--} : ctl) mem
> forward (BinOp e1 op e2) env ctl mem
>   =  xforward e1 env (BinOp1 {--} op e2 env : ctl) mem
> forward (String s) env ctl mem
>   =  xbackward (VString s) ctl mem
> forward Type env ctl mem
>   =  xbackward VType ctl mem
> forward (Var x) env ctl mem
>   =  case lookup x env of
>      Nothing  ->  coreDumped (text "unbound variable: " <> text x)
>      Just v   ->  xbackward v ctl mem
> forward (Let x1 e1 e2) env ctl mem
>   =  xforward e1 env (Let1 x1 {--} e2 env : ctl) mem
> forward (Fun x e) env ctl mem
>   =  xbackward (Closure x e False env) ctl mem
> forward (App e1 e2) env ctl mem
>   =  xforward e1 env (App1 {--} e2 env : ctl) mem

> forward (Rec p e) env ctl mem
>   =  case v of
>      Nothing -> coreDumped (text "only functions or records of functions may be recursively defined:" <> nl <>
>                   (ppExpr (Rec p e)))
>      Just v  -> xbackward v ctl mem
>      where v = knot e (match' p (fromJust v) ++ env)
> forward (Sel e l) env ctl mem
>   =  xforward e env (Sel1 {--} l env : ctl) mem
> forward (Record []) env ctl mem
>   =  xbackward (VRecord []) ctl mem
> forward (Record (f : fs)) env ctl mem
>   =  xforward e env (recordi [] r fs env : ctl) mem
>      where (r, e) = split f

> forward (Con c e) env ctl mem
>   =  xforward e env (Con1 c {--} : ctl) mem
> forward (Seq e1 e2) env ctl mem
>   =  xforward e1 env (Seq1 {--} e2 env : ctl) mem
> forward (Ref e) env ctl mem
>   =  xforward e env (Ref1 {--} : ctl) mem
> forward (Assign e1 e2) env ctl mem
>   =  xforward e1 env (Assign1 {--} e2 env : ctl) mem
> forward (Deref e) env ctl mem
>   =  xforward e env (Deref1 {--} : ctl) mem
> forward (LetCC x e) env ctl mem
>   =  xforward e ((x, Cont ctl) : env) ctl mem
> forward (Case e rs) env ctl mem
>   =  xforward e env (Case1 {--} rs env : ctl) mem
> forward (Continue e1 e2) env ctl mem
>   =  xforward e1 env (Continue1 {--} e2 env : ctl) mem
> forward (Throw e) env ctl mem
>   =  xforward e env (Throw1 {--} : ctl) mem
> forward (Try e rs) env ctl mem
>   =  xforward e env (Try1 {--} rs env : ctl) mem
> --forward (Write e) env ctl mem
> --  =  xforward e env (Write1 {--} : ctl) mem
> forward (Call f e) env ctl mem
>   =  xforward e env (Call1 f {--} : ctl) mem
> forward (While e1 e2) env ctl mem
>   =  xforward e1 env (While1 e1 e2 env : ctl) mem

 ### 
HACK: Eintrag der Form ("", "struct") dient als Marker fr den Anfang
einer neuen hierarchischen Umgebung.

> forward (LetOpen e1 e2) env ctl mem
>   =  xforward e1 env (LetOpen1 {--} e2 env : ctl) mem
> forward (Push e) env ctl mem
>   =  xforward e (("", VString "struct") : env) ctl mem -- HACK
> forward (Pop e) env ctl mem -- HACK
>   =  xforward e (env1 ++ tail env3) ctl mem
>      where (env1, env2) = span (\ (x, v) -> not (null x)) env
>            env3 = dropWhile (\ (x, v) -> not (null x)) (tail env2)
> forward Struct env ctl mem
>   =  xbackward (VStruct (takeWhile (\ (x, v) -> not (null x)) env)) ctl mem
> forward (Lookup e x) env ctl mem
>   =  xforward e env (Lookup1 {--} x : ctl) mem

> forward (Class e) env ctl mem
>   =  xbackward (VClass e env) ctl mem
> forward (New e) env ctl mem
>   =  xforward e env (New1 {--} : ctl) mem
> forward Empty env ctl mem
>   =  xbackward (VObject []) ctl mem
> forward (Append e1 e2) env ctl mem
>   =  xforward e1 env (Append1 {--} e2 env : ctl) mem
> forward (Method x e) env ctl mem
>   =  xbackward (VObject [(x, VMethod e env)]) ctl mem
> forward (Inherit e) env ctl mem
>   =  xforward e env (Inherit1 {--} : ctl) mem
> forward (InheritIn x e1 e2) env ctl mem
>   =  xforward e1 env (InheritIn1 x {--} e2 env : ctl) mem

> forward (ArrFun e1 x e2) env ctl mem
>   =  xforward e1 env (ArrFun1 {--} x e2 env : ctl) mem
> forward (ArrList []) env ctl mem
>   =  xbackward (VArray (listArray (0, -1) [])) ctl mem
> forward (ArrList (e : es)) env ctl mem
>   =  xforward e env (arrListi [] es env : ctl) mem
> forward (Subscript e1 e2) env ctl mem
>   =  xforward e1 env (Subscript1 {--} e2 env : ctl) mem

 ###

Tying the recursive knot (returns `Nothing' if `e' is not a value).

> knot :: Expr -> Env -> Maybe Val

> knot (Fun p e) env
>   =  Just (Closure p e True env)
> knot (Record rs) env
>   =  do  rs' <- mapM (\ r -> knotRow r env) rs
>          return (vrecord rs')

begin{experimentell}
Just an ordinary call-by-name interpreter (denotational semantics) for
a pure lambda-calculus.

> knot (Var x) env
>   =  lookup x env -- mglicherweise ein Blackhole, das Haskell entdeckt
> knot (App e1 e2) env
>   =  do  Closure p e b env' <- knot e1 env
>          v2 <- knot e2 env
>          knot e (match' p v2 ++ env')
> knot (Sel e l) env
>   =  do  VRecord rs <- knot e env
>          select l rs

end{experimentell}

> knot e env
>   =  Nothing
>
> knotRow (Named s e) env
>   =  do  v <- knot e env
>          return (Named s v)
> knotRow (Numbered e) env
>   =  do  v <- knot e env
>          return (Numbered v)

> split (Named l a)  = (Named l (), a)
> split (Numbered a) = (Numbered (), a)

Backward step (return). Here all the dynamic checks are happening.

> backward :: Frame -> Val -> Stack -> Mem -> IO State

> backward (If1 {--} e2 e3 env) (VBool b) ctl mem
>   =  if b then xforward e2 env ctl mem
>           else xforward e3 env ctl mem
> backward (If1 {--} e2 e3 env) v ctl mem
>   =  coreDumped (text "`if' expects a Bool") -- TODO
> backward (UnOp1 op {--}) (VNat v) ctl mem
>   =  xbackward (unop op v) ctl mem
> backward (UnOp1 op {--}) v ctl mem
>   =  coreDumped (text "`" <> pretty op <> text "' expects a natural number:" <> nl <>
>        (group (pretty op <+> ppVal v)))
> backward (BinOp1 {--} op e2 env) v1 ctl mem
>   =  xforward e2 env (BinOp2 v1 op {--} : ctl) mem

NEW: |div| and |mod| raise an exception.

> backward (BinOp2 (VNat v1) Div {--}) (VNat 0) ctl mem
>   =  xbackward (VCon "Div" (VRecord [])) (Throw1 {--} : ctl) mem
> backward (BinOp2 (VNat v1) Mod {--}) (VNat 0) ctl mem
>   =  xbackward (VCon "Mod" (VRecord [])) (Throw1 {--} : ctl) mem

> backward (BinOp2 (VNat v1) op {--}) (VNat v2) ctl mem
>   =  xbackward (binopNat v1 op v2) ctl mem
> backward (BinOp2 (VString v1) op {--}) (VString v2) ctl mem
>   =  xbackward (binopString v1 op v2) ctl mem
> backward (BinOp2 v1 op {--}) v2 ctl mem
>   =  coreDumped (text "`" <> pretty op <> text "' expects two natural numbers or two strings:" <> nl <>
>        (group (ppVal v1 <+> pretty op <+> ppVal v2)))
> backward (Let1 p1 {--} e2 env) v1 ctl mem
>   =  xforward e2 (match' p1 v1 ++ env) ctl mem

> backward (App1 {--} e2 env) v1@(Closure _ _ _ _) ctl mem  -- Funktionsaufruf
>   =  xforward e2 env (App2 v1 {--} : ctl) mem

 backward (App1 {--} e2 env) v1@(VMethod _ _ _) ctl mem    -- Methodenaufruf
   =  xforward e2 env (MApp2 v1 {--} env : ctl) mem

> backward (App1 {--} e2 env) v1 ctl mem
>   =  coreDumped (text "`@': function or method expected:" <> nl <>
>        (group (ppVal v1 <> text " @" <+> ppExpr e2))) -- TODO: env fehlt
> backward (App2 (Closure PatType e _ env) {--}) v2 ctl mem
>   |  case v2 of { VType -> False; _ -> True }  -- insert missing type argument
>   =  xforward e env (TApp2 {--} v2 : ctl) mem
> backward (App2 (Closure p e _ env) {--}) v2 ctl mem
>   =  xforward e (match' p v2 ++ env) ctl mem
> backward (TApp2 {--} v2) (Closure PatType e _ env) ctl mem
>   |  case v2 of { VType -> False; _ -> True }  -- insert missing type argument
>   =  xforward e env (TApp2 {--} v2 : ctl) mem
> backward (TApp2 {--} v2) (Closure p e _ env) ctl mem
>   =  xforward e (match' p v2 ++ env) ctl mem
> backward (TApp2 {--} v2) v1 ctl mem
>   =  coreDumped (text "`@': function or method expected:" <> nl <>
>        (group (ppVal v1 <> text " @" <+> ppVal v2))) -- TODO: env fehlt (??)

 backward (MApp2 (VMethod p e env) {--} env') v2 ctl mem
   =  xforward e (("self", fromJust (lookup "self" env')) : match' p v2 ++ env) ctl mem

Direkter Methodenaufruf in der Klassendefinition (m e) ohne Qualifier
(c.m() oder self.m()): `m e' ist eine Abkrzung fr `self.m e'.

> backward (Sel1 {--} l _) (VRecord v) ctl mem
>   =  case select l v of
>      Nothing -> coreDumped (text "missing field: " <> pretty l)
>      Just v  -> xbackward v ctl mem

> backward (Sel1 {--} (AlNumLab l) _) v@(VObject env) ctl mem
>   =  case lookup l env of
>      Nothing  ->  coreDumped (text "message not understood: " <> text l)
>      Just (VMethod e env') ->  xforward e (("self", v) : env') ctl mem
>--      Just v  -> xbackward v ctl mem

Methoden werden speziell behandelt: `self' wird an das Objekt selbst
gebunden.  In allen anderen Fllen entspricht die Selektion der
Record-Selektion.

> backward (Sel1 {--} (AlNumLab l) env') (VSuper env) ctl mem
>   =  case lookup l env of
>      Nothing  ->  coreDumped (text "message not understood: " <> text l)
>      Just (VMethod e env) ->  xforward e (("self", fromJust (lookup "self" env')) : env) ctl mem
>--      Just (VMethod p e env) ->  xbackward (Closure p e False (("self", ) : env)) ctl mem
>--      Just v  -> xbackward v ctl mem

Bei einem `super'-Aufruf wird nicht das Objekt selbst an `self'
gebunden, sondern die gegenwrtige Bindung von `self' weitergereicht.

> backward (Sel1 {--} l _) v ctl mem
>   =  coreDumped (text "`.': record or object expected:" <> nl <>
>        (group (ppVal v <> text " ." <+> ppLab l)))

 backward (Force1 {--}) (Method e env) ctl mem
   =  xforward e env ctl mem

> backward (Recordi fs1 r1 (f2 : fs2) env) v ctl mem
>   =  xforward e2 env (recordi (fs1 ++ [join r1 v]) r2 fs2 env : ctl) mem
>      where (r2, e2) = split f2
> backward (Recordn fs l) v ctl mem
>   =  xbackward (vrecord (fs ++ [join l v])) ctl mem

---

> backward (Con1 c {--}) v ctl mem
>   =  xbackward (VCon c v) ctl mem

> backward (Seq1 {--} e2 env) _ ctl mem
>   =  xforward e2 env ctl mem
> backward (Ref1 {--}) v ctl (n, mem)
>   =  xbackward (Addr n) ctl (n + 1, (n, v) : mem)
> backward (Assign1 {--} e2 env) v1 ctl mem
>   =  xforward e2 env (Assign2 v1 {--} : ctl) mem
> backward (Assign2 (Addr a1) {--}) v2 ctl (n, mem)
>   =  xbackward (VRecord []) ctl (n, update a1 v2 mem)
> backward (Assign2 v1 {--}) v2 ctl (n, mem)
>   =  coreDumped (text "`:=': address expected") -- TODO

> backward (Deref1 {--}) (Addr a) ctl mem@(n, mem')
>   =  case lookup a mem' of
>      Nothing -> coreDumped (text "segmentation violation: &" <> pretty a)
>      Just v  -> xbackward v ctl mem
> backward (Deref1 {--}) v ctl mem
>   =  coreDumped (text "`!': address expected") -- TODO

> backward (Case1 {--} rs env) v ctl mem
>   =  case msum [ match p v >>= \ env -> return (env, e) | (p, e) <- rs ] of
> --     Nothing        -> coreDumped (text "missing case") -- TODO
>      Nothing        -> xbackward (VCon "Match" (VRecord [])) (Throw1 {--} : ctl) mem
>      Just (env', e) -> xforward e (env' ++ env) ctl mem
> backward (Continue1 {--} e2 env) v1 ctl mem
>   =  xforward e2 env (Continue2 v1 {--} : ctl) mem
> backward (Continue2 (Cont ctl1) {--}) v2 ctl mem
>   =  xbackward v2 ctl1 mem
> backward (Throw1 {--}) v (Try1 {--} rs env : ctl) mem
>   =  case msum [ match p v >>= \ env -> return (env, e) | (p, e) <- rs ] of
>      Nothing        -> xbackward v (Throw1 {--} : ctl) mem
>      Just (env', e) -> xforward e (env' ++ env) ctl mem
> backward (Throw1 {--}) (VCon c v1) (_ : ctl) mem
> --  =  backward (Throw1 {--}) (VCon c v1) ctl mem
>   =  xbackward (VCon c v1) (Throw1 {--} : ctl) mem
> backward (Throw1 {--}) (VCon c v1) [] mem
>   =  coreDumped (text "uncaught exception: " <> ppVal (VCon c v1))
> backward (Throw1 {--}) v [] mem
>   =  runtimeTypeError "throw" "an exception" v
> backward (Try1 {--} rs env) v ctl mem
>   =  xbackward v ctl mem

TODO: Fehlerbehandlung Call1 (a la contracts?!)

> backward (Call1 Chr {--}) (VNat n) ctl mem
>   =  xbackward (VString [chr (fromInteger n)]) ctl mem
> backward (Call1 Chr {--}) v ctl mem
>   =  runtimeTypeError "chr" "a natural number" v
> backward (Call1 GetChar {--}) (VRecord []) ctl mem
>   =  getChar >>= \ c -> xbackward (VString [c]) ctl mem
> backward (Call1 GetChar {--}) v ctl mem
>   =  runtimeTypeError "getchar" "a unit" v

> backward (Call1 Head {--}) (VString s) ctl mem
>   =  xbackward (VString (case s of "" -> ""; c : _ -> [c])) ctl mem
> backward (Call1 Head {--}) v ctl mem
>   =  runtimeTypeError "head" "a string" v
> backward (Call1 Null {--}) (VString s) ctl mem
>   =  xbackward (VBool (null s)) ctl mem
> backward (Call1 Null {--}) v ctl mem
>   =  runtimeTypeError "null" "a string" v

> backward (Call1 Ord {--}) (VString (c : _)) ctl mem
>   =  xbackward (VNat (toInteger (ord c))) ctl mem
> backward (Call1 Ord {--}) v ctl mem
>   =  runtimeTypeError "ord" "a string" v
> backward (Call1 PutChar {--}) (VString (c : _)) ctl mem
>   =  putChar c >> xbackward (VRecord []) ctl mem
> backward (Call1 PutChar {--}) v ctl mem
>   =  runtimeTypeError "putchar" "a string" v
> backward (Call1 Show {--}) v ctl mem
>   =  xbackward (VString (render OneLine (ppVal v))) ctl mem

> backward (Call1 Tail {--}) (VString s) ctl mem
>   =  xbackward (VString (case s of "" -> ""; _ : s' -> s')) ctl mem
> backward (Call1 Tail {--}) v ctl mem
>   =  runtimeTypeError "tail" "a string" v

> backward (Call1 Write {--}) (VString s) ctl mem
>   =  putStr s >> xbackward (VRecord []) ctl mem
> backward (Call1 Write {--}) v ctl mem
>   =  runtimeTypeError "write" "a string" v
> backward (Call1 Size {--}) (VArray a) ctl mem
>   =  xbackward (VNat (snd (bounds a) + 1)) ctl mem
> backward (Call1 Size {--}) v ctl mem
>   =  runtimeTypeError "size" "an array" v

> backward (While1 e1 e2 env) (VBool b) ctl mem
>   =  if b then xforward e2 env (Seq1 {--} (While e1 e2) env : ctl) mem
>           else xbackward (VRecord []) ctl mem

> backward (LetOpen1 {--} e2 env) (VStruct env') ctl mem
>   =  xforward e2 (env' ++ env) ctl mem
> backward (Lookup1 {--} x) (VStruct env) ctl mem
>   =  forward (Var x) env ctl mem

 backward (LetInherit1 Nothing {--} e2 env) (VObject env') ctl mem
   =  xforward e2 (env' ++ env) ctl mem
 backward (LetInherit1 (Just x) {--} e2 env) (VObject env') ctl mem
   =  xforward e2 ((x, VSuper env') : env' ++ env) ctl mem

Wird die geerbte Klasse an einen (Werte-) Bezeichner gebunden, wird
die Klasse als `Super' markiert in der Umgebung abgelegt, sozusagen
als Unterobjekt.

> backward (New1 {--}) (VClass e env) ctl mem
>   =  xforward e env ctl mem
> backward (Append1 {--} e2 env) v1 ctl mem
>   =  xforward e2 env (Append2 v1 {--} : ctl) mem
> backward (Append2 (VObject env1) {--}) (VObject env2) ctl mem
>   =  xbackward (VObject (env2 ++ env1)) ctl mem
> backward (Inherit1 {--}) (VClass e env) ctl mem
>   =  xforward e env ctl mem
> backward (InheritIn1 x {--} e2 env) (VClass e env') ctl mem
>   =  xforward e env' (InheritIn2 x {--} e2 env : ctl) mem
> backward (InheritIn2 x {--} e2 env) v@(VObject env') ctl mem
>   =  xforward e2 ((x, VSuper env') : env) (InheritIn3 x v {--} : ctl) mem
> backward (InheritIn3 x (VObject env1) {--}) (VObject env2) ctl mem
>   =  xbackward (VObject (env2 ++ env1)) ctl mem


> backward (ArrFun1 {--} x e2 env) (VNat n) ctl mem
>   |  n == 0
>   =  xbackward (VArray (listArray (0, -1) [])) ctl mem
>   |  otherwise
>   =  xforward e2 ((x, VNat 0) : env) (arrFun2i 1 n [] x e2 env : ctl) mem
> backward (ArrFun2i i n vs x e2 env) v ctl mem
>   =  xforward e2 ((x, VNat i) : env) (arrFun2i (i + 1) n (vs ++ [v]) x e2 env : ctl) mem
> backward (ArrListi vs (e : es) env) v ctl mem
>   =  xforward e env (arrListi (vs ++ [v]) es env : ctl) mem
> backward (ArrListn vs) v ctl mem
>   =  xbackward (VArray (listArray (0, genericLength vs) (vs ++ [v]))) ctl mem
> backward (Subscript1 {--} e2 env) v1 ctl mem
>   =  xforward e2 env (Subscript2 v1 {--} : ctl) mem
> backward (Subscript2 (VArray a) {--}) (VNat i) ctl mem
>   |  inRange (bounds a) i
>   =  xbackward (a!i) ctl mem
>   |  otherwise
>   =  xbackward (VCon "Sub" (VRecord [])) (Throw1 {--} : ctl) mem
> backward (Subscript2 _ {--}) _ ctl mem
>   =  coreDumped (text "array subscript expects an array and a Nat")

TODO: Wenn der Musterabgleich fehlschlgt: Match exception auslsen.
Als Zwischenloesung geben wir die leere Bindungsliste zurueck. Das
fuehrt dann normalerweise zu einem Core-Dump wegen einer ungebundenen
Variable, das ist aber immer noch besser als ein Haskell-Runtime-Error.

> match' :: Pat -> Val -> [(Var,Val)]
> match' p v = maybe [] id (match p v)

> match :: Pat -> Val -> Maybe [(Var,Val)]
> match PatFail v
>   =  Nothing
> match PatSucceed v
>   =  Just []
> match (PatVar x) v
>   =  Just [(x, v)]
> match (PatCon c p) (VCon c' v)
>   |  c == c'
>   =  match p v
> match PatType VType
>   =  Just []
> match (PatCon c p) v
>   =  Nothing
> match (PatRecord ps) (VRecord vs)
>   =  do  envs <- matchRow ps vs
>          return (Prelude.concat envs)

> match (PatBoth p1 p2) v
>   =  do  env1 <- match p1 v
>          env2 <- match p2 v
>          return (env1 ++ env2)
> match (PatAlt p1 p2) v
>   =  match p1 v `mplus` match p2 v
> match _ _
>   =  Nothing

Annahme: Records sind sortiert/normiert.

> matchRow [] _         =  return []
> matchRow _ []         =  return []  -- eigentlich: Nothing
> matchRow ps@(Named x p : pr) (Named y v : vr)
>   |  x == y           =  liftM2 (:) (match p v) (matchRow pr vr)
> matchRow ps@(Numbered p : pr) (Numbered v : vr)
>                       =  liftM2 (:) (match p v) (matchRow pr vr)
> matchRow pr (_ : vr)  =  matchRow pr vr

> join (Named l _)  a = Named l a
> join (Numbered _) a = Numbered a

> select (AlNumLab l) = sel
>   where sel []                 =  Nothing
>         sel (Named l' a : fs)
>           | l == l'            =  Just a
>           | otherwise          =  sel fs
>         sel (Numbered a : fs)  =  sel fs
> select (NumLab i)                      = sel (i - 1) -- Numerierung beginnt bei 1 (wg `fst' und `snd')
>   where sel n (Named l' a : fs)        =  sel n fs
>         sel 0 (Numbered a : fs)        =  Just a
>         sel (n + 1) (Numbered a : fs)  =  sel n fs
>         sel _ _                        =  Nothing

> update a v [] = []
> update a v ((b, w) : bs)
>   | a == b    = (b, v) : bs
>   | otherwise = (b, w) : update a v bs

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Primitive operations}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

Unary and binary operators.

> --unop Neg n  =  VNat (negate n) -- Hmm: fr natrliche Zahlen
> unop Neg n  =  VNat n

> binopNat :: Integer -> BinOp -> Integer -> Val
> binopNat n1 Add n2  =  VNat  (n1 +  n2)
> binopNat n1 Sub n2  =  VNat  ((n1 -  n2) `max` 0) -- spezielle Subtraktion auf den natrlichen Zahlen
> binopNat n1 Mul n2  =  VNat  (n1 *  n2)
> binopNat n1 Div n2  =  VNat  (n1 `div` n2)
> binopNat n1 Mod n2  =  VNat  (n1 `mod` n2)
> binopNat n1 Lt  n2  =  VBool (n1 <  n2)
> binopNat n1 Leq n2  =  VBool (n1 <= n2)
> binopNat n1 Eq  n2  =  VBool (n1 == n2)
> binopNat n1 Neq n2  =  VBool (n1 /= n2)
> binopNat n1 Geq n2  =  VBool (n1 >= n2)
> binopNat n1 Gt  n2  =  VBool (n1 >  n2)

NB. `div' und 'mod' sind die sogenannten mathematischen Varianten (im
Unterschied zu `quot' und `rem'): (a + km) `mod` m = a `mod` m.

TODO: hier knnen Fehler auftreten: Division durch 0!

> binopString :: String -> BinOp -> String -> Val
> binopString s1 Cat s2  =  VString (s1 ++ s2)
> binopString s1 Add s2  =  VString (s1 ++ s2)
> binopString s1 Lt  s2  =  VBool (s1 <  s2)
> binopString s1 Leq s2  =  VBool (s1 <= s2)
> binopString s1 Eq  s2  =  VBool (s1 == s2)
> binopString s1 Neq s2  =  VBool (s1 /= s2)
> binopString s1 Geq s2  =  VBool (s1 >= s2)
> binopString s1 Gt  s2  =  VBool (s1 >  s2)

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Initial environment}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

Die initiale Umgebung dient dazu, primitive Konstanten und Funktionen
wie `false' oder `write' bekanntzumachen, da diese nicht in
`System.bpl' definiert werden knnen. Alternativ: Bezeichner als
Schlsselwrter reservieren, so dass `write' stets die primitive
Funktion meint.

> closure x e
>   =  Closure (PatVar x) e False []

> stdenv', stdenv :: Env
> stdenv'
>   =  [("show",    closure "x" (Call Show (Var "x"))),
>       ("write",   closure "x" (Call Write (Var "x")))]
> stdenv
>   =  [("chr",     closure "x" (Call Chr (Var "x"))),
>       ("false",   VBool False),
>       ("getchar", closure "x" (Call GetChar (Var "x"))),
>       ("head",    closure "x" (Call Head (Var "x"))),
>       ("null",    closure "x" (Call Null (Var "x"))),
>       ("ord",     closure "x" (Call Ord (Var "x"))),
>       ("putchar", closure "x" (Call PutChar (Var "x"))),
>       ("show",    closure "x" (Call Show (Var "x"))),
>       ("size",    closure "x" (Call Size (Var "x"))),
>       ("tail",    closure "x" (Call Tail (Var "x"))),
>       ("true",    VBool True),
>       ("write",   closure "x" (Call Write (Var "x")))]

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Execute}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> execute :: Options -> Expr -> Env -> Mem -> IO State
> execute options e env mem
>   =  do catch
>           (do hSetBuffering stdout NoBuffering
>               hSetBuffering stdin  NoBuffering)
>           (\_ -> return ())
>         if trace options then
>           debug (Forward e env [] mem)
>          else
>           exec (Forward e env [] mem)

> exec :: State -> IO State
> exec cfg@(Stop v mem)
> --  =  putStrLn (render (Page 80) (ppState cfg)) >> 
>   =  return cfg
> exec cfg@(CoreDumped s)
> --  =  putStrLn (render (Page 80) (ppState cfg)) >> 
>   =  return cfg
> exec cfg 
>   =  step cfg >>= exec

 initial options e
   =  if Bare `elem` options
        then Forward e [] [] (0, [])
        else Forward e stdenv [] (0, [])

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Tracer}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

 trace options e 
   = do hSetBuffering stdout NoBuffering
        hSetBuffering stdin  NoBuffering
        debug (initial options e)
 
> debug cfg
>   =  do  putStrLn (render (Page 80) (ppState cfg))

          putStr "? "
          inp <- getLine

>          inp <- Readline.readline "? "
>          case inp of
>            Nothing -> return cfg
>            Just x ->
>              if x `elem` [""] then
>                step cfg >>= debug
>              else if all isDigit x then
>                apply (read x) step cfg >>= debug
>              else if x `elem` ["q", "quit"] then
>                return cfg
>              else
>                debug cfg

> ppState (Stop v mem) 
>   =  ppVal v
> --     <> nl <> text "mem: " <> block 5 (ppMem mem)
> ppState (Forward  e env ctl mem) 
>   =  text ">>>> " <> block 5 (ppExpr e) <> nl <>
>      text "env: " <> block 5 (ppEnv env) <> nl <>
>      text "ctl: " <> block 5 (ppStack ctl)
>      <> nl <> text "mem: " <> block 5 (ppMem mem)
> ppState (Backward v ctl mem) 
>   =  text "<<<< " <> block 5 (ppVal v) <> nl <>
>      text "ctl: " <> block 5 (ppStack ctl)
>      <> nl <> text "mem: " <> block 5 (ppMem mem)
> ppState (CoreDumped d)  
>   =  text "core dumped: " <> d

> ppStack
>   =  ppList ppFrame

> apply 0 f x
>   =  return x
> apply (n + 1) f x
>   =  apply n f x >>= f

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Tests}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

 cfg = initial (BinOp (Nat 4711) Add (Nat 815))

debug cfg

 go e = debug (initial e)

> fun x e = Fun (PatVar x) e

go (fun "x" (Var "x"))
go (fun "x" (Var "x") $ Nat 4711)
go dd
go (factorial $ Num 3)
go (fromto $ Nat 1 $ Nat 9)

> dd = fun "x" (Var "x" $ Var "x") $ fun "x" (Var "x" $ Var "x") -- ;-)

> factorial
>   = Rec (PatVar "factorial") (fun "n" (
>       If (BinOp (Var "n") Eq (Nat 0)) 
>          (Nat 1) 
>          (BinOp (Var "n") Mul (Var "factorial" $ BinOp (Var "n") Sub (Nat 1))) ))

> fromto
>   = Rec (PatVar "fromto") (fun "l" (fun "r" (
>       If (BinOp (Var "l") Gt (Var "r"))
>          (Con "Nil" (Record []))
>          (Con "Cons" (Record [ Named hd (Var "l"),
>                                Named tl ( Var "fromto" $ BinOp (Var "l") Add (Nat 1) $ Var "r")])) )))

> hd = "head"
> tl = "tail"